//*******************************************************
//
//	Color LCD Clock
//
//	ATmega1284p
//		efuse:0xFD	BOD 2.7V
//		hfuse:0xD1	JTAG禁止, EEPROM消去保護
//		lfuse:0xE2	内蔵CR発振 8MHz
//
//	Compiler:	AVR-GCC		4.3.4
//				AVR-LIBC	1.6.7
//				binutils	2.20
//
//	2011.5.24	KAWAKAMI Yukio
//*******************************************************

#include	<avr/io.h>
#include	<avr/interrupt.h>
#include	<avr/pgmspace.h>
#include	<avr/eeprom.h>
#include	<stdio.h>
#include	<stdlib.h>
#include	<string.h>

#include	"main.h"
#include	"lcd.h"
#include	"diskio.h"
#include	"pff.h"
#include	"com.h"

#include	<util/delay.h>

#define		FIRMWARE_FNAME	"CLCLOCK.BIN"
#define		BACK_FNAME		"CLKBACK.GRP"

FATFS	FatFs;

COM_WORK	Com0;
COM_WORK	Com1;

volatile uchar	Timer;
volatile char	C1Hz;
volatile uchar	BacklightTimer;

void disk_timerproc(void);	// diskio.c

//*******************************************************
//	割り込み停止
//*******************************************************
void stop_interrupt(void){
	uchar	r;

	cli();
	WDR;
	
	// 割り込み禁止
	UCSR0B = 0;
	TIMSK0 = 0;
	EIMSK = 0;

	// 割り込みフラグクリア
	UCSR0A = (1<<TXC0);
	r = UDR0;
	TIFR0 = 0xFF;
	EIFR |= (1<<INTF1);
}


//*******************************************************
//	タイマー０比較一致Ａ割り込み
//	100Hz
//*******************************************************
ISR(TIMER0_COMPA_vect){

	if (Timer)		--Timer;

	disk_timerproc();
}


//*******************************************************
//	INT1割り込み
//	RTCからの正確な 1Hz割り込み
//*******************************************************
ISR(INT1_vect){
	uint	count;

	// CR発振クロック補正。起動してから安定するまで 10秒くらいみておくこと
	// OSCCALが大雑把なのでこの程度で十分
	count = TCNT1L;
	count |= ((uint)TCNT1H) << 8;
	TCNT1H = 0;
	TCNT1L = 0;

	if (count > ((F_CPU + 128) / 256)){
		--OSCCAL;
	} else if (count < ((F_CPU + 128) / 256)){
		OSCCAL++;
	}

	if (BacklightTimer){
		if (--BacklightTimer == 0){
			PORTD &= ~(1<<BR2);
			PORTD &= ~(1<<BR1);	// バックライト輝度最小
		}
	}
	
	C1Hz = 1;
}


//*******************************************************
//	メインルーチン
//*******************************************************
int main(void){
	int		i;
	uchar	sw_count = 0;

	cli();

	PORTA = (1<<RSTB)|(1<<LWR)|(1<<LRD)|(1<<LCS);
	DDRA = (1<<RSTB)|(1<<LWR)|(1<<LRD)|(1<<LRES)|(1<<LRS)|(1<<LCS);

	PORTB = (1<<MOSI)|(1<<SDCS)|(1<<MCS);
	DDRB = (1<<SCK)|(1<<MOSI)|(1<<SDCS)|(1<<RDATA)|(1<<RCLK)|(1<<CCE)|(1<<MCS);

	PORTC = 0;
	DDRC = 0xFF;

	PORTD = (1<<PD7)|(1<<TXD)|(1<<RXD);
	DDRD = (1<<BCE)|(1<<BR2)|(1<<BR1)|(1<<RWR);

	if (MCUSR & (1<<WDRF)){
		// ウォッチドッグリセット
	}
	MCUSR = 0;

	WDR;
	WDTCSR = (1<<WDCE)|(1<<WDE);	// WDCEとWDEを1にしてタイマー変更をする
	WDTCSR = (1<<WDE)|(1<<WDP2)|(1<<WDP1)|(1<<WDP0);	// ウォッチドッグタイマーセット 約２秒

	TCCR0A = (1<<WGM01);			// CTC動作
	TCCR0B = (1<<CS02)|(1<<CS00);	// プリスケーラ 1/1024
	OCR0A = F_CPU/1024/100 - 1;		// タイマカウント 100Hz
	TIMSK0 = (1<<OCIE0A);			// 比較一致Ａ割り込み許可

	TCCR1A = 0;
	TCCR1B = (1<<CS12);	// フリーラン, プリスケーラ 1/256
	TCNT1H = 0;
	TCNT1L = 0;
	

	EICRA = (1<<ISC11);			// INT1立ち下がり割り込み
	EIMSK = (1<<INT1);			// INT1割り込み許可

	SMCR = (1<<SE);			// スリープ許可 アイドルモード

	// TWI, ADC 電力カット
	PRR0 = (1<<PRTWI)|(1<<PRADC);
	ACSR = (1<<ACD);			// アナログコンパレータ禁止

	SPCR = (1<<SPE)|(1<<MSTR)|(1<<SPR1);	// SPIマスタ Mode 0
	SPSR = (1<<SPI2X);						// fosc/32 この段階ではロースピード

	for (i = 0; i < 10; i++){
		SPI_in();
	}
	
	sei();

	wait_timer(10);

	LCD_init();

	PORTD |= (1<<BCE);	// バックライト ON
	PORTD |= (1<<BR2);	// バックライト輝度2

	pf_mount(&FatFs);
	{
		ushort	fdate, ftime;
		char	fname[16];
		// ファームウェアファイルのタイムスタンプチェック
		strcpy_P(fname, PSTR(FIRMWARE_FNAME));
		if (get_fdatetime(fname, &fdate, &ftime)){
			if ((eeprom_read_word(EEPROM_FIRM_FDATE) != fdate)||
				(eeprom_read_word(EEPROM_FIRM_FTIME) != ftime)){
				void	(*boot)(void);
				boot = (void (*)(void))(ADRS_BOOT/2);
				// ブートローダにジャンプ
				stop_interrupt();
				(*boot)();
			}
		}
	}

	load_back();					// 背景画像を Flashメモリへ読み込み
	put_grp(ADRS_DATA + 4, 0, 0);	// Flashメモリの背景画像を表示

	PORTD |= (1<<BR1);	// バックライト輝度3

	locate(0, LCD_HEIGHT - 8);
	print_P(PSTR("CL Clock"));
	locate(LCD_WIDTH - 6 * 8, LCD_HEIGHT - 8);
	print_P(PSTR("V."VERSION));

	// FLL安定するまで待つ
	for (i = 0; i < 5; i++){
		C1Hz = 0;
		while(!C1Hz){
			WDR;
			SLEEP;
		}
	}
	
	init_com(&Com0, B9600);
	write_str_P(&Com0, PSTR("Color LCD Clock Version " VERSION "\r\nOK.\r\n"));

	BacklightTimer = BACKLIGHT_TIME;	// バックライトタイマー動作開始

	for(;;){
		char	buff[32];
		struct RTC_TIME	rtc;
		uchar	sw;

		Timer = 1;
		while(Timer){
			WDR;
			SLEEP;
			if (C1Hz)	break;
		}
		
		if (C1Hz){
			C1Hz = 0;
			read_RTC(&rtc);

			sprintf_P(buff, PSTR("%d/%02d/%02d"),
					  rtc.year + 2000, rtc.month, rtc.day);
			locate(0, LCD_HEIGHT - 8);
			print(buff);

			sprintf_P(buff, PSTR("%02d:%02d:%02d"),
					  rtc.hour, rtc.min, rtc.sec);
			locate(LCD_WIDTH - 6 * 8, LCD_HEIGHT - 8);
			print(buff);
		}

		job_serial();

		sw = check_sw();
		if (sw){
			PORTD |= (1<<BR2);
			PORTD |= (1<<BR1);
			BacklightTimer = BACKLIGHT_TIME;
		} else {
			sw_count = 0;
		}
		if (sw == 1){
			if (++sw_count >= 200){		// SW1長押しで時間調整
				BacklightTimer = 0;		// バックライト自動消灯OFF
				set_time();				// マニュアル時間調整
				sw_count = 0;
				BacklightTimer = BACKLIGHT_TIME;
			}
		}

	}
}


//*******************************************************
//	シリアル入力処理
//*******************************************************
void job_serial(void){
	static uchar com_buff[32];
	static uchar com_buff_p = 0;

	while(check_com(&Com0)){
		uchar	c;

		c = read_com(&Com0);
		if (c == LF){
			;
		} else if (c == CR){
			struct RTC_TIME	rtc;
			int	year, month, day;
			int hour, min, sec;

			if (sscanf_P((char *)com_buff,
						 PSTR("%d/%d/%d %d:%d:%d"),
						 &year, &month, &day,
						 &hour, &min, &sec) == 6){
				if ((year >= 2011)&&(year <= 2099)&&
					(month >= 1)&&(month <= 12)&&
					(day >= 1)&&(day <= 31)&&
					(hour >= 0)&&(hour <= 23)&&
					(min >= 0)&&(min <= 59)&&
					(sec >= 0)&&(sec <= 59)){
					rtc.year = year % 100;
					rtc.month = month;
					rtc.day = day;
					rtc.hour = hour;
					rtc.min = min;
					rtc.sec = sec;
					set_week(&rtc);
					set_RTC(&rtc);
					write_str_P(&Com0, PSTR("OK\r\n"));
				}
			} else if (strncasecmp_P((char *)com_buff,
									 PSTR("VER"), 3) == 0){
				write_str_P(&Com0, PSTR("Version " VERSION "\r\n"));
			} else if (strncasecmp_P((char *)com_buff,
									 PSTR("TIME"), 4) == 0){
				struct RTC_TIME	rtc;
				uchar	buff[32];
				read_RTC(&rtc);
				sprintf_P((char *)buff, PSTR("%d/%02d/%02d %02d:%02d:%02d\r\n"),
						  rtc.year + 2000,
						  rtc.month,
						  rtc.day,
						  rtc.hour,
						  rtc.min,
						  rtc.sec);
				write_str(&Com0, buff);
			} else if (strncasecmp_P((char *)com_buff,
									 PSTR("BACKLIGHT "), 10) == 0){
				int n;
				if (strncasecmp_P((char *)&com_buff[10],
								  PSTR("OFF"), 3) == 0){
					PORTD &= ~(1<<BCE);
				} else if (strncasecmp_P((char *)&com_buff[10],
								  PSTR("ON"), 2) == 0){
					PORTD |= (1<<BCE);
				} else if (sscanf_P((char *)&com_buff[10],
								  PSTR("%d"), &n) == 1){
					if ((n >= 0)&&(n <= 3)){
						if (n & 1){
							PORTD |= (1<<BR1);
						} else {
							PORTD &= ~(1<<BR1);
						}
						if (n & 2){
							PORTD |= (1<<BR2);
						} else {
							PORTD &= ~(1<<BR2);
						}
						PORTD |= (1<<BCE);
					}
				}
			} else if (strcasecmp_P((char *)com_buff,
									 PSTR("BACKLIGHT")) == 0){
				uchar	bright = 0;
				uchar	buff[8];

				write_str_P(&Com0, PSTR("BackLight "));
				if (PORTD & (1<<BR1))	bright += 1;
				if (PORTD & (1<<BR2))	bright += 2;
				if (PORTD & (1<<BCE)){
					sprintf_P((char *)buff, PSTR("%d\r\n"), bright);
					write_str(&Com0, buff);
				} else {
					write_str_P(&Com0, PSTR("OFF\r\n"));
				}
			}

			com_buff_p = 0;
			com_buff[0] = 0;
		} else {
			if (com_buff_p < (sizeof(com_buff)-1)){
				com_buff[com_buff_p++] = c;
				com_buff[com_buff_p] = 0;
			}
		}
	}
}


//*******************************************************
//	マニュアル時間合わせ
//*******************************************************
void set_time(void){
	struct RTC_TIME	rtc;
	char	buff[3];
	
	read_RTC(&rtc);
	sprintf_P(buff, PSTR("%02d"), rtc.year);
	locate(6*2, LCD_HEIGHT - 8);
	print(buff);
	sprintf_P(buff, PSTR("%02d"), rtc.month);
	locate(6*5, LCD_HEIGHT - 8);
	print(buff);
	sprintf_P(buff, PSTR("%02d"), rtc.day);
	locate(6*8, LCD_HEIGHT - 8);
	print(buff);

	sprintf_P(buff, PSTR("%02d"), rtc.hour);
	locate(LCD_WIDTH-6*8, LCD_HEIGHT - 8);
	print(buff);
	sprintf_P(buff, PSTR("%02d"), rtc.min);
	locate(LCD_WIDTH-6*5, LCD_HEIGHT - 8);
	print(buff);
	locate(LCD_WIDTH-6*2, LCD_HEIGHT - 8);
	print_P(PSTR("00"));

	// 年～時まではタイムアウトするが、
	if (set_time_sub(6*2, LCD_HEIGHT-8, &(rtc.year), 0, 99))	return;
	if (set_time_sub(6*5, LCD_HEIGHT-8, &(rtc.month), 1, 12))	return;
	if (set_time_sub(6*8, LCD_HEIGHT-8, &(rtc.day), 1, 31))		return;
	if (set_time_sub(LCD_WIDTH-6*8, LCD_HEIGHT-8, &(rtc.hour), 0, 23))	return;
	// 分の所まで来たらタイムアウトしない
	while (set_time_sub(LCD_WIDTH-6*5, LCD_HEIGHT-8, &(rtc.min), 0, 59));

	// 曜日設定
	set_week(&rtc);
	rtc.sec = 0;	// 00秒
	set_RTC(&rtc);	// RTC設定

	// SWが OFFになるまで待つ
	while(check_sw()){
		wait_timer(1);
	}

	wait_timer(10);

	return;
}


//****************************************************************
//	マニュアル時間合わせサブ
//****************************************************************
uchar set_time_sub(uchar x, uchar y, uchar *n, uchar min, uchar max){
	char	buff[3];
	uchar	data;
	uchar	timeout;
	uchar	repeat;
	uchar	sw;

	// SWが OFFになるまで待つ
	while(check_sw()){
		wait_timer(1);
	}

	data = *n;
	repeat = 0;
	for (;;){
		WDR;
		sprintf_P(buff, PSTR("%02d"), data);
		locate(x, y);
		print(buff);
		Timer = 100;
		timeout = 0;
		while(!(sw = check_sw())){
			WDR;
			SLEEP;
			locate(x, y);
			if ((Timer == 20)||(Timer == 70)){
				print_P(PSTR("\x7F\x7F"));
			} else if (Timer == 50){
				print(buff);
			} else if (Timer == 0){
				if (++timeout >= 15)	return 1;	// 15秒放置でタイムアウト
				Timer = 100;
				print(buff);
			}
			repeat = 0;
		}
		if (sw == 1){
			locate(x, y);
			print(buff);
			*n = data;
			return 0;
		} else if (sw == 2){
			if (repeat == 0){
				if (++data >= max){
					data = min;
				}
				repeat++;
			} else if (repeat == 100){
				if (++data > max){
					data = min;
				}
				repeat = 50;
			} else {
				repeat++;
			}
			wait_timer(1);
		}
	}
}


//****************************************************************
//	背景画像ファイルデータを Flashメモリへ読み込み
//****************************************************************
void load_back(void){
	char	fname[16];
	ushort	fdate, ftime;
	ulong	adrs;
	WORD	n;
	FRESULT	res;
	char	msg[8];

	strcpy_P(fname, PSTR(BACK_FNAME));
	if (!get_fdatetime(fname, &fdate, &ftime))	return;
	if ((pgm_read_word(ADRS_DATA) == ftime)&&
		(pgm_read_word(ADRS_DATA + 2) == fdate))	return;

	if (pf_open(fname) != FR_OK)	return;

	print_P(PSTR("Loading " BACK_FNAME "\n"));

	adrs = ADRS_DATA;
	write_flash(adrs++, ftime);
	write_flash(adrs++, ftime >> 8);
	write_flash(adrs++, fdate);
	write_flash(adrs++, fdate >> 8);

	while(FatFs.fptr < FatFs.fsize){
		uchar	buff[32];
		uchar	i;

		WDR;

		if (((adrs - 4) & 0x00FF) == 0){
			ltoa(adrs - 4, msg, 16);
			strupr(msg);
			print(msg);
			putch(CR);
		}

		res = pf_read(buff, sizeof(buff), &n);
		if (res != FR_OK)	break;

		i = 0;
		while(n){
			write_flash(adrs++, buff[i++]);
			--n;
		}
	}

	// 半端な分を書き込むためにパディング
	while((adrs & (SPM_PAGESIZE-1))){
		write_flash(adrs++, 0xFF);
	}

	ltoa(adrs - 4, msg, 16);
	strupr(msg);
	print(msg);
	putch(LF);

	if (res != FR_OK){
		print_P(PSTR("Can't Loading All\n"));
		wait_timer(150);
	}
	wait_timer(100);
}


//****************************************************************
//	MCUの Flash ROMにデータ書き込み
//	実際の書き込みは 1ページ分のデータがそろった時点で行われるので
//	半端がある場合は 0xFFでも送って数を揃える。
//	また、adrsの管理はここではしていないので外でやること。
//****************************************************************
void write_flash(ulong adrs, uchar data){
	static uchar buff[SPM_PAGESIZE];
	uint	p;

	p = (uint)(adrs & (SPM_PAGESIZE-1));
	buff[p] = data;

	if (p == (SPM_PAGESIZE-1)){
		// ブートローダエリアの関数呼び出し
		if (adrs < ADRS_LIMIT){		// 書き込み上限
			void	(*self_program)(ulong, ushort *);
			self_program = (void (*)(ulong, ushort *))(ADRS_WRITE_FLASH/2);
			cli();
			(*self_program)(adrs - p, (ushort *)buff);
			sei();
		}
	}
}
	

//****************************************************************
//	Flashメモリ上のデータ読み込み
//****************************************************************
void pgm_read_block(ulong adrs, uchar *buff, uint size){

	while(size){
		*buff++ = pgm_read_byte_far(adrs++);
		--size;
	}
}


//****************************************************************
//	Flashメモリ上のGPRデータ表示
//****************************************************************
void put_grp(ulong adrs, uchar x, uchar y){
	GRPHEADER	grp_header;
	int		width, height;
	short	size;			// 128x160=20480なので shortでOK
	ushort	comp1, comp2;

	// ファイルヘッダ読み込み
	pgm_read_block(adrs, (uchar *)&grp_header, sizeof(grp_header));
	adrs += sizeof(grp_header);

	// ファイルヘッダチェック
	if (grp_header.mark != 0x5047)	return;

	// ファイルフォーマットチェック
	width = grp_header.width;
	height = grp_header.height;
	comp1 = grp_header.comp1;
	comp2 = grp_header.comp2;
	if ((grp_header.type != 1)||
		(width > LCD_WIDTH)||
		(height > LCD_HEIGHT)){
		return;
	}

	// 画面内に収まるか
	if ((x + width > LCD_WIDTH)||
		(y + height > LCD_HEIGHT)){
		return;
	}

	// 液晶データウィンドウ
	set_window(x, y, x + width - 1, y + height - 1);

	size = width;
	size *= height;

	PORTA &= ~(1<<LCS);

	while(size > 0){
		ushort	data;;

		data = pgm_read_word_far(adrs);
		adrs += 2;

		if (comp1 && (comp1 == data)){	// ランレングス圧縮マーク
			ushort	repeat;

			repeat = pgm_read_word_far(adrs);
			adrs += 2;
			data = pgm_read_word_far(adrs);
			adrs += 2;
			size -= repeat;
			while(repeat){
				PORTA &= ~(1<<LWR);
				PORTC = data >> 8;
				PORTA |= (1<<LWR);
				--repeat;
				PORTA &= ~(1<<LWR);
				PORTC = data;
				PORTA |= (1<<LWR);
			}
		} else {
			PORTA &= ~(1<<LWR);
			PORTC = data >> 8;
			PORTA |= (1<<LWR);
			--size;
			PORTA &= ~(1<<LWR);
			PORTC = data;
			PORTA |= (1<<LWR);
		}
	}

	PORTA |= (1<<LCS);
}


//****************************************************************
//	ファイルの更新時間取得
//****************************************************************
char get_fdatetime(char *fname, ushort *fdate, ushort *ftime){
	FRESULT	res;
	FILINFO	finfo;
	DIR		dir;
	char	path[16];

	path[0] = 0;
	res = pf_opendir(&dir, path);
	if (res != FR_OK)	return 0;
	for(;;){
		res = pf_readdir(&dir, &finfo);
		if (res != FR_OK || finfo.fname[0] == 0)	return 0;
		if (!(finfo.fattrib & AM_DIR)){
			if (strcmp(finfo.fname, fname) == 0){
				*fdate = finfo.fdate;
				*ftime = finfo.ftime;
				return 1;
			}
		}
	}
}


//*******************************************************
//	GRPファイル表示
//*******************************************************
char load_grp(char *fname, uchar x, uchar y){
	GRPHEADER	grp_header;
	FRESULT	res;
	WORD	n;
	int		width, height;
	long	size;
	char	ret = 0;
	uchar	remain = 0;
	ushort	buff[32], *p;
	ushort	comp1, comp2;
	ushort	repeat = 0;


	// ファイルオープン
	if ((res = pf_open(fname) != FR_OK))	return res;	// ファイルオープンエラー

	// ファイルヘッダ読み込み
	pf_read((void *)&grp_header, sizeof(grp_header), &n);

	// ファイルヘッダチェック
	if (grp_header.mark != 0x5047){ ret = 20; goto EXIT; }	// ファイルタイプエラー

	// ファイルフォーマットチェック
	width = grp_header.width;
	height = grp_header.height;
	comp1 = grp_header.comp1;
	comp2 = grp_header.comp2;
	if ((grp_header.type != 1)||
		(width > LCD_WIDTH)||
		(height > LCD_HEIGHT)){
		ret = 30;
		goto EXIT;
	}

	// 画面内に収まるか
	if ((x + width > LCD_WIDTH)||
		(y + height > LCD_HEIGHT)){
		ret = 40;
		goto EXIT;
	}

	// 液晶データウィンドウ
	set_window(x, y, x + width - 1, y + height - 1);

	size = width;
	size *= height;

	PORTA &= ~(1<<LCS);

	p = buff;
	while(size > 0){
		ushort	data;;

		WDR;

		if (remain == 0){
			pf_read((void *)buff, sizeof(buff), &n);
			if (n <= 0){ ret = 50; goto EXIT; }	// 読み込みエラー
			remain = n / sizeof(ushort);
			p = buff;
		}
		data = *p++;
		--remain;

		if (comp1 && (comp1 == data)){	// ランレングス圧縮マーク
			if (remain == 0){
				pf_read((void *)buff, sizeof(buff), &n);
				if (n <= 0){ ret = 50; goto EXIT; }	// 読み込みエラー
				remain = n / sizeof(ushort);
				p = buff;
			}
			repeat = *p++;
			--remain;
			if (remain == 0){
				pf_read((void *)buff, sizeof(buff), &n);
				if (n <= 0){ ret = 50; goto EXIT; }	// 読み込みエラー
				remain = n / sizeof(ushort);
				p = buff;
			}
			data = *p++;
			--remain;
			while(repeat){
				PORTA &= ~(1<<LWR);
				PORTC = data >> 8;
				PORTA |= (1<<LWR);
				--size;
				PORTA &= ~(1<<LWR);
				PORTC = data;
				PORTA |= (1<<LWR);
				--repeat;
			}
		} else {
			PORTA &= ~(1<<LWR);
			PORTC = data >> 8;
			PORTA |= (1<<LWR);
			--size;
			PORTA &= ~(1<<LWR);
			PORTC = data;
			PORTA |= (1<<LWR);
			--repeat;
		}
	}

  EXIT:
	PORTA |= (1<<LCS);

	return ret;
}

//*******************************************************
//	スイッチ入力チェック
//*******************************************************
uchar check_sw(void){
	uchar	sw;

	DDRC = 0;
	PORTC = 0xFF;

	_delay_us(1);

	sw = ~PINC;

	PORTC = 0;
	DDRC = 0xFF;
	
	if (sw & 1)	return 1;
	if (sw & 2)	return 2;

	return 0;
}


//*******************************************************
//	Clock moduleから時間読み込み
//*******************************************************
void read_RTC(struct RTC_TIME *rtc){
	char	bit;
	uchar	c, i;
	uchar	buff[13];

	DDRB &= ~(1<<RDATA);

	PORTD &= ~(1<<RWR);
	PORTB |= (1<<CCE);

	// 4bit x 13
	for (i = 0; i < 13; i++){
		c = 0;
		for (bit = 0; bit < 4; bit++){
			NOP;	// RTC-4543の 3V時の tCLKH, tCLKL は 0.75μs以上であること
			NOP;
			NOP;
			NOP;
			c >>= 1;
			PORTB |= (1<<RCLK);
			NOP;
			NOP;
			NOP;
			NOP;
			if (PINB & (1<<RDATA)){
				c |= 0x08;
			}
			PORTB &= ~(1<<RCLK);
		}
		buff[i] = c;
	}

	PORTB &= ~(1<<CCE);

	if (buff[SEC_HIGH] & 0x08){		// FDTビット(電源電圧検出ビット)
		buff[SEC_HIGH] &= 0x07;
		buff[DAY_LOW] = 1;
		buff[DAY_HIGH] = 0;
		buff[MONTH_LOW] = 1;
		buff[MONTH_HIGH] = 0;
		buff[YEAR_LOW] = 0;
		buff[YEAR_HIGH] = 0;
	}
	buff[MIN_HIGH] &= 0x07;
	buff[HOUR_HIGH] &= 0x03;
	buff[WEEK] &= 0x07;
	buff[DAY_HIGH] &= 0x03;
	buff[MONTH_HIGH] &= 0x01;

	rtc->sec = (buff[SEC_HIGH] * 10) + buff[SEC_LOW];
	rtc->min = (buff[MIN_HIGH] * 10) + buff[MIN_LOW];
	rtc->hour = (buff[HOUR_HIGH] * 10) + buff[HOUR_LOW];
	rtc->week = buff[WEEK];
	rtc->min &= 0x7F;
	rtc->day = (buff[DAY_HIGH] * 10) + buff[DAY_LOW];
	rtc->month = (buff[MONTH_HIGH] * 10) + buff[MONTH_LOW];
	rtc->year = (buff[YEAR_HIGH] * 10) + buff[YEAR_LOW];
}


//*******************************************************
//	RTC時計合わせ
//*******************************************************
void set_RTC(struct RTC_TIME *rtc){
	uchar	buff[13];
	uchar	i, bit, c;

	buff[SEC_LOW] = rtc->sec % 10;
	buff[SEC_HIGH] = rtc->sec / 10;
	buff[MIN_LOW] = rtc->min % 10;
	buff[MIN_HIGH] = rtc->min / 10;
	buff[HOUR_LOW] = rtc->hour % 10;
	buff[HOUR_HIGH] = rtc->hour / 10;
	buff[WEEK] = rtc->week;
	buff[DAY_LOW] = rtc->day % 10;
	buff[DAY_HIGH] = rtc->day / 10;
	buff[MONTH_LOW] = rtc->month % 10;
	buff[MONTH_HIGH] = rtc->month / 10;
	buff[YEAR_LOW] = rtc->year % 10;
	buff[YEAR_HIGH] = rtc->year / 10;

	DDRB |= (1<<RDATA);
	
	PORTD |= (1<<RWR);
	PORTB |= (1<<CCE);

	// 4bit x 13
	for (i = 0; i < 13; i++){
		c = buff[i];
		for (bit = 0; bit < 4; bit++){
			if (c & 1){
				PORTB |= (1<<RDATA);
			} else {
				PORTB &= ~(1<<RDATA);
			}
			NOP;	// RTC-4543の 3V時の tCLKH, tCLKL は 0.75μs以上であること
			NOP;
			NOP;
			NOP;
			PORTB |= (1<<RCLK);
			NOP;
			NOP;
			NOP;
			NOP;
			NOP;
			c >>= 1;
			PORTB &= ~(1<<RCLK);
		}
	}

	PORTB &= ~(1<<CCE);
}


//*******************************************************
//	曜日計算をして rtc.weekにセット
//*******************************************************
void set_week(struct RTC_TIME *rtc){
	long	y, m, d, days;

	y = rtc->year + 2000;
	m = rtc->month;
	d = rtc->day;

	if (m < 3){
		--y;
		m += 12;
	}
	days = 365 * y + y / 4 - y / 100 + y / 400 + 306 * (m + 1) / 10 + d - 428;
	rtc->week = days % 7;
}


PROGMEM char MonthDays[]
	= { 0, 31, 28, 31, 30, 31, 30, 31, 31, 30, 31, 30, 31 };

//*******************************************************
//	秒数加算
//*******************************************************
void add_sec(struct RTC_TIME *rtc, uchar sec){

	rtc->min += sec / 60;
	rtc->sec += sec % 60;
	if (rtc->sec >= 60){
		rtc->sec -= 60;
		if (++rtc->min >= 60){
			rtc->min -= 60;
			if (++rtc->hour >= 24){
				uchar	last_day;
				if (++rtc->week >= 7){
					rtc->week = 0;
				}
				rtc->hour = 0;
				last_day = pgm_read_byte(&MonthDays[rtc->month]);
				if ((rtc->month == 2)&&((rtc->year % 4) == 0)){
					last_day++;
				}
				if (++rtc->day > last_day){
					rtc->day = 1;
					if (++rtc->month >= 13){
						rtc->month = 1;
						++rtc->year;
					}
				}
			}
		}
	}
}


//*******************************************************
//	SPI入力
//*******************************************************
uchar SPI_in(void){

	SPDR = 0xFF;
	while(!(SPSR & (1<<SPIF)));
	return SPDR;
}


//*******************************************************
//	SPI出力
//*******************************************************
void SPI_out(uchar data){

	SPDR = data;
	while(!(SPSR & (1<<SPIF)));
}


//*******************************************************
//	シリアルRAM読み込み(N25S830HA)
//*******************************************************
uchar read_ram(uint adrs){
	uchar	data;

	PORTB &= ~(1<<MCS);
	SPI_out(0x03);
	SPI_out(adrs >> 8);
	SPI_out((uchar)adrs);
	data = SPI_in();
	PORTB |= (1<<MCS);

	return data;
}


//*******************************************************
//	シリアルRAMブロック読み込み(N25S830HA)
//*******************************************************
void read_ram_block(uint adrs, uchar *buff, int size){

	PORTB &= ~(1<<MCS);
	SPI_out(0x01);
	SPI_out(0x41);	// Burst Mode
	PORTB |= (1<<MCS);
	
	PORTB &= ~(1<<MCS);
	SPI_out(0x03);
	SPI_out(adrs >> 8);
	SPI_out((uchar)adrs);
	while(size){
		*buff++ = SPI_in();
		--size;
	}
	PORTB |= (1<<MCS);

	PORTC &= ~(1<<MCS);
	SPI_out(0x01);
	SPI_out(0x01);	// Word Mode
	PORTC |= (1<<MCS);
}


//*******************************************************
//	シリアルRAM書き込み(N25S830HA)
//*******************************************************
void write_ram(uint adrs, uchar data){

	PORTB &= ~(1<<MCS);
	SPI_out(0x02);
	SPI_out(adrs >> 8);
	SPI_out((uchar)adrs);
	SPI_out(data);
	PORTB |= (1<<MCS);
}


//*******************************************************
//	シリアルRAMブロック書き込み(N25S830HA)
//*******************************************************
void write_ram_block(uint adrs, uchar *buff, int size){

	PORTB &= ~(1<<MCS);
	SPI_out(0x01);
	SPI_out(0x41);	// Burst Mode
	PORTB |= (1<<MCS);
	
	PORTB &= ~(1<<MCS);
	SPI_out(0x02);
	SPI_out(adrs >> 8);
	SPI_out((uchar)adrs);
	while(size){
		SPI_out(*buff++);
		--size;
	}
	PORTB |= (1<<MCS);

	PORTB &= ~(1<<MCS);
	SPI_out(0x01);
	SPI_out(0x01);	// Word Mode
	PORTB |= (1<<MCS);
}


//*******************************************************
//	10ms単位時間待ち
//*******************************************************
void wait_timer(uchar tm){

	Timer = tm;

	while(Timer){
		WDR;
		SLEEP;
	}
}
